/*
 * eval.c
 * functions for builtin calculator
 *
 * (created: Massimiliano Ghilardi (Cosmos), Jan 15th, 1995)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <limits.h>

#include "powwow.h"
#include "commands.h"
#include "list.h"
#include "mapping.h"
#include "tty.h"
#include "eval.h"

#ifdef USE_RANDOM
#  define get_random random
#  define init_random srandom
#else
#  define get_random lrand48
#  define init_random srand48
#endif

char *first_valid();

char *error_msg[] = {
      "unknown error",
      "math stack overflow",
      "math stack underflow",
      "expression syntax",
      "operator expected",
      "value expected",
      "division by zero",
      "operand or index out of range",
      "missing right parenthesis",
      "missing left parenthesis",
      "out of memory",
      "operator is not (yet) implemented",
      "infinite loop",
      "numeric value expected",
      "string expected",
      "stack overflow",
      "stack underflow",
      "missing label",
      "missing separator `;'",
      "attempted to execute recursive #history",
      "user break",
      "too many defined variables",
      "undefined variable",
      "invalid digit in numeric value",
      "bad attribute syntax",
      "invalid variable name",
};

operator_list op_list[] = {
    { 0, 0,     0,	  ""     },     /* null operator */
      
    { 1, LEFT,  BINARY,     ","    },
      
    { 2, RIGHT, BINARY,     "="    },
    { 2, RIGHT, BINARY,     "||="  },
    { 2, RIGHT, BINARY,     "^^="  },
    { 2, RIGHT, BINARY,     "&&="  },
    { 2, RIGHT, BINARY,     "|="   },
    { 2, RIGHT, BINARY,     "^="   },
    { 2, RIGHT, BINARY,     "&="   },
    { 2, RIGHT, BINARY,     "<<="  },
    { 2, RIGHT, BINARY,     ">>="  },
    { 2, RIGHT, BINARY,     "+="   },
    { 2, RIGHT, BINARY,     "-="   },
    { 2, RIGHT, BINARY,     "*="   },
    { 2, RIGHT, BINARY,     "/="   },
    { 2, RIGHT, BINARY,     "%="   },
      
    { 3, LEFT,  BINARY,     "||"   },
      
    { 4, LEFT,  BINARY,     "^^"   },
      
    { 5, LEFT,  BINARY,     "&&"   },
      
    { 6, LEFT,  BINARY,     "|"    },
      
    { 7, LEFT,  BINARY,     "^"    },
      
    { 8, LEFT,  BINARY,     "&"    },
      
    { 9, LEFT,  BINARY,     "<"    },
    { 9, LEFT,  BINARY,     "<="   },
    { 9, LEFT,  BINARY,     ">"    },
    { 9, LEFT,  BINARY,     ">="   },
    { 9, LEFT,  BINARY,     "=="   },
    { 9, LEFT,  BINARY,     "!="   },
      
    {10, LEFT,  BINARY,     "<<"   },
    {10, LEFT,  BINARY,     ">>"   },
      
    {11, LEFT,  BINARY,     "+"    },
    {11, LEFT,  BINARY,     "-"    },
      
    {12, LEFT,  BINARY,     "*"    },
    {12, LEFT,  BINARY,     "/"    },
    {12, LEFT,  BINARY,     "%"    },
      
    {14, LEFT,  BINARY,     ":<"   },
    {14, LEFT,  BINARY,     ":>"   },
    {14, LEFT,  BINARY,     "<:"   },
    {14, LEFT,  BINARY,     ">:"   },
    {14, LEFT,  BINARY,     ".<"   },
    {14, LEFT,  BINARY,     ".>"   },
    {14, LEFT,  BINARY,     "<."   },
    {14, LEFT,  BINARY,     ">."   },
    {14, LEFT,  BINARY,     ":"    },
    {14, LEFT,  BINARY,     "."    },
    {14, LEFT,  BINARY,     "?"    },
      
    { 0, 0,     0,	  ""     },     /* null operator */
      
    { 0, RIGHT, PRE_UNARY,  "("    },
    { 0, RIGHT, POST_UNARY, ")"    },
      
    {13, RIGHT, PRE_UNARY,  "!"    },
    {13, RIGHT, PRE_UNARY,  "~"    },
    {13, RIGHT, PRE_UNARY,  "++"   },
    {13, RIGHT, POST_UNARY, "++"   },
    {13, RIGHT, PRE_UNARY,  "--"   },
    {13, RIGHT, POST_UNARY, "--"   },
    {13, RIGHT, PRE_UNARY,  "*"    },
    {13, RIGHT, PRE_UNARY,  "%"    },
    {13, RIGHT, PRE_UNARY,  "rand" },
    {13, RIGHT, PRE_UNARY,  "attr" },
      
    {14, LEFT,  PRE_UNARY,  ":?"   },
    {14, LEFT,  PRE_UNARY,  ".?"   },
      
    {15, RIGHT, PRE_UNARY,  "+"    },
    {15, RIGHT, PRE_UNARY,  "-"    },
    {15, RIGHT, PRE_UNARY,  "@"    },
    {15, RIGHT, PRE_UNARY,  "$"    },
      
    { 0, 0,     PRE_UNARY,  ""     },    /* null operator */
    { 0, 0,     POST_UNARY, ""     }     /* null operator */
};

stack stk;
char *line;
int depth;
int error;

int push_op();
int pop_op();
int push_obj();
int pop_obj();
int check_operator();
int check_object();
void check_delete();
int exe_op();
int whichfirst();
int compare_and_unload();
int _eval();

void print_error(err_num)
int err_num;
{
    tty_clear_input_line(1);
    if (error == NO_MEM_ERROR) {
	tty_printf("#system call error: %s (%d", "malloc", ENOMEM);
	tty_printf(": %s)\n", strerror(ENOMEM));
    } else
	tty_printf("#error: %s.\n", error_msg[err_num]);
}

int push_op(op)
operator *op;
{
    if (stk.curr_op<MAX_STACK) {
	stk.op[++stk.curr_op]=*op;
	return 1;
    }
    else {
	print_error(error=STACK_OV_ERROR);
	return 0;
    }
}

int pop_op(op)
operator *op;
{
    if (stk.curr_op>=0) {
	*op=stk.op[stk.curr_op--];
	return 1;
    }
    else {
	print_error(error=STACK_UND_ERROR);
	return 0;
    }
}

int push_obj(obj)
object *obj;
{
    object *tmp;
    
    int curr=stk.curr_obj;
    
    if (curr<MAX_STACK) {
	tmp = stk.obj + (stk.curr_obj = ++curr);
	bcopy(obj, tmp, sizeof(object));
	return 1;
    }
    else {
	print_error(error=STACK_OV_ERROR);
	return 0;
    }
}

int pop_obj(obj)
object *obj;
{
    object *tmp;
    
    int curr=stk.curr_obj;
    
    if (curr>=0) {
	tmp = stk.obj + curr;
	stk.curr_obj--;
	bcopy(tmp, obj, sizeof(object));
	return 1;
    }
    else {
	print_error(error=STACK_UND_ERROR);
	return 0;
    }
}

int check_operator(side, op, mindepth)
char side; operator *op; int mindepth;
{
    int i, max, match, len;
    char *name, c, d;
    
    if (!(c=*line) || c == CMDSEP) {
	*op = side==BINARY ?
	    (operator)null
	    : side==LEFT ?
	    (operator)pre_null
	    : (operator)post_null;
	
	return 1;
    }
    else if ((c=='$' || c=='@') && (d=line[1]) && (isalpha(d) || d=='_'))
	return 0;           /* Danger! found named variable */
    
    else if (side==LEFT && c=='(') {
	line++;
	depth++;
	*op=(operator)left_paren;
	return 1;
    }
    else if (side==RIGHT && c==')') {
	if (--depth >= mindepth) {
	    line++;
	    *op=(operator)right_paren;
	}
	else 		   /* exit without touching the parenthesis */
	    *op=(operator)post_null;
	return 1;
    }
    
    for(max=match=0, i=(side==BINARY ? 1 : LOWEST_UNARY_CODE);
	*(name = op_list[i].name); i++)
	if ((len=strlen(name)) > max &&
	    (side==BINARY || side==op_list[i].syntax) &&
	    !strncmp(line, name, (size_t)len)) {
	    match=i;
	    max=len;
	}
    
    if (match) {
	*op=(operator)match;
	line+=max;
	return 1;
    }
    else {
	*op= side==BINARY ? null : side==PRE_UNARY ? pre_null : post_null;
	if (side==BINARY)
	    print_error(error=NO_OPERATOR_ERROR);
    }
    return 0;
}

int check_object(obj)
object *obj;
{
    long i=0, base = 10;
    char c, *end, digit;
    
    if(c=*line, c == '#' || isdigit(c)) {
	while (c == '#' || isalnum(c)) {
	    digit = isdigit(c);
	    if (c == '#') {
		base = i;
		i = 0;
		if (!base)
		    base = 16;
	    } else {
		i *= base;
		if (digit)
		    i += (c - '0');
		else {
		    if (c >= 'a' && c <= 'z')
			c = (c - 'a') + 'A';
		    if (c - 'A' + 10 >= base) {
			print_error(error=OUT_BASE_ERROR);
			return 0;
		    }
		    i += (c - 'A' + 10);
		}
	    }
	    c=*++line;
	}
	obj->type=TYPE_NUM;
	obj->num=i;
	i=1;
    }
    else if(c=='\"') {
	end=first_valid(++line, '\"');
	if (*end) {
	    obj->type=TYPE_TXT;
	    obj->txt=ptrmcpy(obj->txt, line, end-line);
	    if (error)
		print_error(error);
	    else {
		ptrunescape(obj->txt);
		i=1;
		line=end+1;
	    }
	}
    }
    else if ((c=='$' || c=='@') && (c=line[1]) && (isalpha(c) || c=='_')) {
	varnode *named_var;                       /* Found named variable */
	
	if (*(line++) == '@') {
	    i = 0;
	    obj->type = TYPE_NUM_VAR;
	}
	else {
	    i = 1;
	    obj->type = TYPE_TXT_VAR;
	}
	end = line + 1;
	while ((c=*end) && (isalpha(c) || c=='_' || isdigit(c)))
	    end++;
	c = *end; *end = '\0';
	if (!(named_var = *lookup_varnode(line, (char)i))) {
	    named_var = add_varnode(line, (char)i);
	    if (error) {
		print_error(error);
		return 0;
	    }
	    if (echo_int) {
		PRINT_INT ("#new variable: %s\n", line - 1);
	    }
	}
	*end = c;
	line = end;
	obj->num = named_var->index;
	i = 1;
    }
    else if (!strncmp(line, "timer", 5)) {
	obj->type = TYPE_NUM;
	obj->num = diff_vtime(&now, &ref_time);
	line += 5;
	i = 1;
    }
    else if (!strncmp(line, "map", 3)) {
	char buf[MAXMAPLEN + 1];
	map_sprintf(buf);
	obj->type = TYPE_TXT;
	obj->txt = ptrmcpy(obj->txt, buf, strlen(buf));
	if (error)
	    print_error(error);
	else {
	    line += 3;
	    i = 1;
	}
    }
    else if (!strncmp(line, "noattr", 6)) {
	obj->type = TYPE_TXT;
	obj->txt = ptrmcat(obj->txt, tty_modestandoff, strlen(tty_modestandoff));
	obj->txt = ptrmcat(obj->txt, tty_modenorm, strlen(tty_modenorm));
	if (error)
	    print_error(error);
	else {
	    line += 6;
	    i = 1;
	}
    }
    else
	print_error(error=NO_VALUE_ERROR);
    
    return (int)i;
}

void check_delete(obj)
object *obj;
{
    if (obj->type==TYPE_TXT && obj->txt)
	ptrdel(obj->txt);
}

int exe_op(op)
operator *op;
{
    object o1, o2, *p=NULL;
    long *l, rnd, delta;
    ptr src = NULL, dst = NULL, start = NULL;
    int srclen;
    char *ssrc, *tmp;
    int ret=0, i=0, j=0, danger=0;
    
    o1.txt = o2.txt = NULL;
    
    switch (*op) {
     case (operator)comma:
	if (pop_obj(&o2) && pop_obj(&o1));
	else if (error) return 0;
	p=&o2;
	ret=1;
	break;
     case (operator)eq:
	if (pop_obj(&o2) && pop_obj(&o1));
	else if (error) return 0;
	
	if (o2.type==TYPE_NUM_VAR) {
	    o2.num = *VAR[o2.num].num;
	    o2.type = TYPE_NUM;
	}
	
	if (o1.type==TYPE_NUM_VAR && o2.type==TYPE_NUM) {
	    *VAR[o1.num].num = o2.num;
	    p=&o2;
	    ret=1;
	}
	else if (o1.type==TYPE_TXT_VAR &&
		 (o2.type==TYPE_TXT || o2.type==TYPE_TXT_VAR)) {
	    
	    if (o2.type==TYPE_TXT_VAR) {
		o2.txt = ptrdup(*VAR[o2.num].str);
		if (error) break;
		o2.type=TYPE_TXT;
	    }
	    
	    *VAR[o1.num].str = ptrcpy(*VAR[o1.num].str, o2.txt);
	    if (error) break;
	    p=&o2;
	    ret=1;
	}
	else {
	    check_delete(&o1);
	    check_delete(&o2);
	    error=SYNTAX_ERROR;
	}
	
	break;
     case (operator)or_or_eq:
     case (operator)xor_xor_eq:
     case (operator)and_and_eq:
     case (operator)or_eq:
     case (operator)xor_eq:
     case (operator)and_eq:
     case (operator)lshift_eq:
     case (operator)rshift_eq:
     case (operator)plus_eq:
     case (operator)minus_eq:
     case (operator)times_eq:
     case (operator)div_eq:
     case (operator)ampersand_eq:
	if (pop_obj(&o2) && pop_obj(&o1));
	else if (error) return 0;
	
	if (o2.type==TYPE_NUM_VAR) {
	    o2.num = *VAR[o2.num].num;
	    o2.type = TYPE_NUM;
	}
	
	if (o1.type==TYPE_NUM_VAR && o2.type==TYPE_NUM) {
	    l=VAR[o1.num].num;
	    
	    switch (*op) {
	     case (operator)or_or_eq:  if ( o2.num) *l = 1;   break;
	     case (operator)xor_xor_eq:if ( o2.num) *l = !*l; break;
	     case (operator)and_and_eq:if (!o2.num) *l = 0;   break;
	     case (operator)or_eq:     *l  |= o2.num; break;
	     case (operator)xor_eq:    *l  ^= o2.num; break;
	     case (operator)and_eq:    *l  &= o2.num; break;
	     case (operator)lshift_eq: *l <<= o2.num; break;
	     case (operator)rshift_eq: *l >>= o2.num; break;
	     case (operator)plus_eq:   *l  += o2.num; break;
	     case (operator)minus_eq:  *l  -= o2.num; break;
	     case (operator)times_eq:  *l  *= o2.num; break;
	     case (operator)div_eq:    *l  /= o2.num; break;
	     case (operator)ampersand_eq:
		if ((*l %= o2.num) < 0) *l += o2.num; break;
	    }
	    o2.num=*l;
	    p=&o2;
	    ret=1;
	}
	else if (*op==(operator)plus_eq && o1.type==TYPE_TXT_VAR &&
		 (o2.type==TYPE_TXT || o2.type==TYPE_TXT_VAR)) {
	    
	    if (o2.type==TYPE_TXT)
		src=o2.txt;
	    else
		src=*VAR[o2.num].str;
	    
	    *VAR[o1.num].str =
		ptrcat(*VAR[o1.num].str, src);
	    check_delete(&o2);
	    
	    /* trick: ptrcpy(NULL, ...) allocates a ptr
	     * and copies into it */
	    dst = ptrcpy(dst, *VAR[o1.num].str);
	    if (error) break;

	    o1.type=TYPE_TXT;
	    o1.txt=dst;
	    p=&o1;
	    ret=1;
	}
	else {
	    check_delete(&o1);
	    check_delete(&o2);
	    error=SYNTAX_ERROR;
	    break;
	}
	break;
     case (operator)or_or:
     case (operator)xor_xor:
     case (operator)and_and:
     case (operator)or:
     case (operator)xor:
     case (operator)and:
     case (operator)less:
     case (operator)less_eq:
     case (operator)greater:
     case (operator)greater_eq:
     case (operator)eq_eq:
     case (operator)not_eq:
     case (operator)lshift:
     case (operator)rshift:
     case (operator)minus:
     case (operator)plus:
     case (operator)times:
     case (operator)division:
     case (operator)ampersand:
	if (pop_obj(&o2) && pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_NUM_VAR) {
	    o1.num = *VAR[o1.num].num;
	    o1.type = TYPE_NUM;
	}
	if (o2.type==TYPE_NUM_VAR) {
	    o2.num = *VAR[o2.num].num;
	    o2.type = TYPE_NUM;
	}
	
	if (o1.type==TYPE_NUM && o2.type==TYPE_NUM) {
	    if (!o2.num &&
		(*op==(operator)division || *op==(operator)ampersand)) {
		error=DIV_BY_ZERO_ERROR;
		break;
	    }
	    
	    switch (*op) {
	     case (operator)less:      o1.num = o1.num <  o2.num ? 1 : 0; break;
	     case (operator)less_eq:   o1.num = o1.num <= o2.num ? 1 : 0; break;
	     case (operator)greater:   o1.num = o1.num >  o2.num ? 1 : 0; break;
	     case (operator)greater_eq:o1.num = o1.num >= o2.num ? 1 : 0; break;
	     case (operator)eq_eq:     o1.num = o1.num == o2.num ? 1 : 0; break;
	     case (operator)not_eq:    o1.num = o1.num != o2.num ? 1 : 0; break;
	     case (operator)or_or:     o1.num = o1.num || o2.num; break;
	     case (operator)xor_xor:if (o2.num) o1.num = !o1.num; break;
	     case (operator)and_and:   o1.num = o1.num && o2.num; break;
	     case (operator)or:  o1.num |= o2.num; break;
	     case (operator)xor: o1.num ^= o2.num; break;
	     case (operator)and: o1.num &= o2.num; break;
	     case (operator)lshift:    o1.num <<= o2.num; break;
	     case (operator)rshift:    o1.num >>= o2.num; break;
	     case (operator)minus:   o1.num -= o2.num; break;
	     case (operator)plus:    o1.num += o2.num; break;
	     case (operator)times:   o1.num *= o2.num; break;
	     case (operator)division:o1.num /= o2.num; break;
	     case (operator)ampersand:
		if ((o1.num %= o2.num) < 0) o1.num += o2.num; break;
	    }
	    
	    p=&o1;
	    ret=1;
	}
	else if ((o1.type==TYPE_TXT || o1.type==TYPE_TXT_VAR) &&
		 (o2.type==TYPE_TXT || o2.type==TYPE_TXT_VAR)) {
	    
	    if (o1.type==TYPE_TXT)
		dst = o1.txt;
	    else {
		dst = ptrcpy(dst, *VAR[o1.num].str);
		if (error) break;
		o1.txt = dst;
	    }
	    if (o2.type==TYPE_TXT)
		src=o2.txt;
	    else
		src=*VAR[o2.num].str;
	    
	    p=&o1;
	    ret=1;
	    o1.type = TYPE_NUM;
	    o1.num = ptrcmp(dst, src);
	    
	    switch (*op) {
	      case (operator)plus:	 dst = ptrcat(dst, src);
				         o1.type = TYPE_TXT;   break;
	      case (operator)minus:			       break;
	      case (operator)less:       o1.num = o1.num <  0; break;
	      case (operator)less_eq:    o1.num = o1.num <= 0; break;
	      case (operator)greater:    o1.num = o1.num >  0; break;
	      case (operator)greater_eq: o1.num = o1.num >= 0; break;
	      case (operator)eq_eq:      o1.num = o1.num == 0; break;
	      case (operator)not_eq:     o1.num = o1.num != 0; break;
	      default:
		error=SYNTAX_ERROR;
		p=NULL; ret=0; break;
	    }
	    if (error) {
		p=NULL; ret=0;
		break;
	    }
	    if (o1.type != TYPE_TXT)
		check_delete(&o1);
	    check_delete(&o2);
	}
	else {
	    check_delete(&o1);
	    check_delete(&o2);
	    error=SYNTAX_ERROR;
	    break;
	}
	break;
     case (operator)colon_less:
     case (operator)colon_greater:
     case (operator)less_colon:
     case (operator)greater_colon:
     case (operator)colon:
     case (operator)point_less:
     case (operator)point_greater:
     case (operator)less_point:
     case (operator)greater_point:
     case (operator)point:
	if (pop_obj(&o2) && pop_obj(&o1));
	else if (error) return 0;
	
	if (o2.type==TYPE_NUM_VAR) {
	    o2.num = *VAR[o2.num].num;
	    o2.type = TYPE_NUM;
	}
	
	if ((o1.type!=TYPE_TXT_VAR && o1.type!=TYPE_TXT) || o2.type!=TYPE_NUM) {
	    check_delete(&o1);
	    check_delete(&o2);
	    error=SYNTAX_ERROR;
	    break;
	}
	
	if (o2.num<=0) {
	    check_delete(&o1);
	    error=OUT_RANGE_ERROR;
	    break;
	}
	
	if (o1.type==TYPE_TXT_VAR) {
	    o1.type=TYPE_TXT;
	    o1.txt=dst=NULL;
	    src=start=*VAR[o1.num].str;
	}
	else {
	    /* Potentially dangerous: src and dst are overlapping */
	    src=dst=start=o1.txt;
	    danger=1;
	}
	
	srclen = ptrlen(src);
	ssrc = ptrdata(src);
	
	switch (*op) {
	 case (operator)colon_less:
	    while (o2.num && srclen) {
		/* skip span of multiple word delimeters */
		while (srclen && memchr(DELIM, *ssrc, DELIM_LEN))
		    srclen--, ssrc++, j++;
		/* skip whole words */
		if (srclen && (tmp = memchrs(ssrc, srclen, DELIM, DELIM_LEN)))
		    i=tmp-ssrc, o2.num--, ssrc+=i, j+=i, srclen-=i;
		else break;
	    }
	    
	    if (o2.num) { /* end of valid string before the n-th word */
		if (danger)
		    ;
		else
		    dst = ptrcpy(dst, start);
	    } else {
		if (danger)
		    ptrtrunc(dst, j);
		else
		    dst = ptrmcpy(dst, ptrdata(start), j);
	    }
	    break;
	 case (operator)colon:
	 case (operator)colon_greater:
	    o2.num--;
	    /* skip span of multiple word delimeters */
	    while (srclen && memchr(DELIM, *ssrc, DELIM_LEN))
		srclen--, ssrc++;
	    while (o2.num && srclen) {
		/* skip whole words */
		if (srclen && (tmp = memchrs(ssrc, srclen, DELIM, DELIM_LEN))) {
		    i=tmp-ssrc, o2.num--, ssrc+=i, srclen-=i;
		    /* skip span of multiple word delimeters */
		    while (srclen && memchr(DELIM, *ssrc, DELIM_LEN))
			srclen--, ssrc++;
		} else break;
	    }

	    if (o2.num)  /* end of valid string before the n-th word */
		ptrzero(dst);
	    else {
		if (*op==(operator)colon &&
		    (tmp = memchrs(ssrc, srclen, DELIM, DELIM_LEN))) {
		    dst = ptrmcpy(dst, ssrc, tmp-ssrc);
		}
		else
		    dst = ptrmcpy(dst, ssrc, srclen);
	    }
	    break;
	 case (operator)less_colon:
	    o2.num--;
	    while (o2.num && srclen) {
		/* skip span of multiple word delimeters */
		while (srclen >= 0 && memchr(DELIM, ssrc[srclen], DELIM_LEN))
		    srclen--;
		/* skip whole words */
		if (srclen>=0 && (tmp=memrchrs(ssrc, srclen, DELIM, DELIM_LEN)))
		    o2.num--, srclen=tmp-ssrc;
		else break;
	    }
	    
	    if (o2.num) /* end of valid string before the n-th word */
		ptrzero(dst);
	    else
		dst = ptrmcpy(dst, ssrc, srclen);
	    break;
	 case (operator)greater_colon:
	    o2.num--;
	    while (o2.num && srclen) {
		/* skip span of multiple word delimeters */
		while (srclen >= 0 && memchr(DELIM, ssrc[srclen], DELIM_LEN))
		    srclen--;
		/* skip whole words */
		if (srclen>=0 && (tmp=memrchrs(ssrc, srclen, DELIM, DELIM_LEN)))
		    o2.num--, srclen=tmp-ssrc;
		else break;
	    }
	    
	    if (o2.num) /* end of valid string before the n-th word */
		dst = ptrcpy(dst, start);
	    else
		dst = ptrmcpy(dst, ssrc+srclen+1,
			      ssrc+srclen+1 - ptrdata(start) + ptrlen(start));
	    break;
	 case (operator)point:
	    dst = ptrmcpy(dst, ssrc+o2.num-1, 1);
	    break;
	 case (operator)point_less:
	    j = o2.num < srclen ? o2.num : srclen;
	    if (danger)
		ptrtrunc(dst, j);
	    else
		dst = ptrmcpy(dst, ssrc, j);
	    break;
	 case (operator)less_point:
	    j = srclen-o2.num+1;
	    if (j < 0)
		j = 0;
	    if (danger)
		ptrtrunc(dst, j);
	    else
		dst = ptrmcpy(dst, ssrc, j);
	    break;
	 case (operator)point_greater:
	    j = o2.num-1 < srclen ? o2.num-1 : srclen;
	    dst = ptrmcpy(dst, ssrc+j, srclen-j);
	    break;
	 case (operator)greater_point:
	    j = srclen-o2.num;
	    if (j < 0)
		j = 0;
	    dst = ptrmcpy(dst, ssrc+j, srclen-j);
	    break;
	}
	check_delete(&o2);
	o1.txt = dst;
	if (error)
	    check_delete(&o1);
	else {
	    p=&o1; ret=1;
	}
	break;
     case (operator)colon_question:
     case (operator)point_question:
	if (pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_TXT)
	    src=o1.txt;
	else if (o1.type==TYPE_TXT_VAR)
	    src=*VAR[o1.num].str;
	else {
	    error=SYNTAX_ERROR;
	    break;
	}
	ssrc = ptrdata(src);
	srclen = ptrlen(src);
	
	if (*op==(operator)colon_question) {
	    o1.num = 0;
	    /* skip span of multiple word delimeters */
	    while (srclen >= 0 && memchr(DELIM, *ssrc, DELIM_LEN))
		ssrc++, srclen--;
	    while (srclen) {
		/* skip whole words */
		if (srclen && (tmp=memchrs(ssrc, srclen, DELIM, DELIM_LEN))) {
		    i=tmp-ssrc, o1.num++, ssrc+=i, srclen-=i;
		    /* skip span of multiple word delimeters */
		    while (srclen && memchr(DELIM, *ssrc, DELIM_LEN))
			srclen--, ssrc++;
		} else {
		    srclen=0;
		    o1.num++;
		}
	    }
	}
	else
	    o1.num=srclen;
	
	check_delete(&o1);
	o1.type=TYPE_NUM;
	p=&o1;
	ret=1;
	break;
     case (operator)question:
	if (pop_obj(&o2) && pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_TXT)
	    dst = o1.txt;
	else if (o1.type==TYPE_TXT_VAR)
	    dst = *VAR[o1.num].str;
	else 
	    error = SYNTAX_ERROR;
	
	if (o2.type==TYPE_TXT)
	    src = o2.txt;
	else if (o2.type==TYPE_TXT_VAR)
	    src = *VAR[o2.num].str;
	else 
	    error = SYNTAX_ERROR;
	
	if (!error) {
	    if ((ssrc = ptrptr(dst, src)))
		i = (int)(ssrc - ptrdata(dst)) + 1;
	    else
		i = 0;
	    o1.type = TYPE_NUM;
	    o1.num = i;
	    p=&o1; ret=1;
	}
	check_delete(&o1);
	check_delete(&o2);
	break;
     case (operator)null:
     case (operator)another_null:
	if (pop_obj(&o2) && pop_obj(&o1));
	else if (error) return 0;
	
	check_delete(&o1);
	check_delete(&o2);
	
	o1.type=0, o1.num=0, o1.txt=NULL;
	
	p=&o1;
	ret=1;
	break;
     case (operator)left_paren:
	error=MISSING_PAREN_ERROR;
	break;
     case (operator)right_paren:
	if (pop_op(op));
	else if (error) return 0;
	
	if (*op!=(operator)left_paren)
	    error=MISMATCH_PAREN_ERROR;
	else
	    ret=1;
	
	break;
     case (operator)_random_:
#ifdef NO_RANDOM
	break;
#endif
     case (operator)pre_plus:
     case (operator)pre_minus:
     case (operator)not:
     case (operator)tilde:
	
	if (pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_NUM_VAR) {
	    o1.num = *VAR[o1.num].num;
	    o1.type = TYPE_NUM;
	}
	if (o1.type==TYPE_NUM) {
	    if (*op==(operator)pre_minus)
		o1.num=-o1.num;
	    else if (*op==(operator)not)
		o1.num=!o1.num;
	    else if (*op==(operator)tilde)
		o1.num=~o1.num;
#ifndef NO_RANDOM
	    else if (*op==(operator)_random_) {
		if (o1.num <= 0) {
		    error=OUT_RANGE_ERROR;
		    break;
		} else {
		    delta = LONG_MAX - LONG_MAX % o1.num;
		    while (rnd = get_random(), rnd > delta);
		    /* skip numbers that would alterate distribution */
		    o1.num = rnd / (delta / o1.num);
		}
	    }
#endif
	    p=&o1;
	    ret=1;
	}
	else {
	    check_delete(&o1);
	    error=SYNTAX_ERROR;
	}
	break;
     case (operator)_attr_:
	if (pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_TXT_VAR) {
	    o1.txt = ptrdup(*VAR[o1.num].str);
	    if (error) break;
	    o1.type = TYPE_TXT;
	}     
	
	if (o1.type==TYPE_TXT) {
	    char dummy[CAPLEN]; /* just because attr_string must write somewhere */
	    
	    i = parse_attributes(o1.txt); 
	    if (i == -1) {
		check_delete(&o1);
		error=BAD_ATTR_ERROR;
	    }
	    else {
		attr_string(i, o1.txt, dummy);
		p=&o1;
		ret = 1;
	    }
	}
	else
	    error=NO_STRING_ERROR;
	break;
	
     case (operator)star:
     case (operator)print:
	if (pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_NUM_VAR)
	    o1.num = *VAR[o1.num].num;
	else if (o1.type==TYPE_TXT_VAR)
	    o1.txt = *VAR[o1.num].str;
	
	if (o1.type==TYPE_NUM || o1.type==TYPE_NUM_VAR) {
	    o1.txt = NULL;
	    if (*op==(operator)print) {
		char buf[PARAMLEN];
		/* safe, a long cannot be PARAMLEN chars */
		sprintf(buf, "%ld", o1.num);
		o1.txt = ptrmcpy(o1.txt, buf, strlen(buf));
	    } else {
		char buf = (char)o1.num;
		/* can't fail, PARAMLEN > 1 */
		o1.txt = ptrmcpy(o1.txt, &buf, 1);
	    }
	    if (error) break;
	    o1.type = TYPE_TXT;
	    p=&o1; ret=1;
	}
	else if (o1.type==TYPE_TXT || o1.type==TYPE_TXT_VAR) {
	    if (*op==(operator)print) {
		o1.txt = ptrpad(o1.txt, 1);
		if (error) break;

		ptrdata(o1.txt)[ptrlen(o1.txt)-1] = '\0';
		o1.num = atol(ptrdata(o1.txt));
	    } else {
		if (o1.txt && ptrlen(o1.txt))
		    o1.num = (long)(unsigned char)*ptrdata(o1.txt);
		else
		    o1.num = 0;
	    }
	    check_delete(&o1);
	    o1.type = TYPE_NUM;
	    p=&o1; ret=1;
	}
	else 
	    error=SYNTAX_ERROR;
	break;
     case (operator)pre_plus_plus:
     case (operator)post_plus_plus:
     case (operator)pre_minus_minus:
     case (operator)post_minus_minus:
	if (pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_NUM_VAR) {
	    l=VAR[o1.num].num;
	    o1.type=TYPE_NUM;
	    
	    if (*op==(operator)pre_plus_plus)
		o1.num=++*l;
	    else if (*op==(operator)post_plus_plus)
		o1.num=(*l)++;
	    else if (*op==(operator)pre_minus_minus)
		o1.num=--*l;
	    else
		o1.num=(*l)--;
	    
	    p=&o1;
	    ret=1;
	}
	else {
	    check_delete(&o1);
	    error=SYNTAX_ERROR;
	}
	break;
     case (operator)a_circle:
	if (pop_obj(&o1));
	else if (error) return 0;
	
	if (o1.type==TYPE_NUM_VAR) {
	    o1.type=TYPE_NUM;
	    o1.num=*VAR[o1.num].num;
	}
	
	if (o1.type==TYPE_NUM) {
	    if (o1.num<-NUMVAR || o1.num>=NUMPARAM) {
		error=OUT_RANGE_ERROR;
		break;
	    }
	    o1.type=TYPE_NUM_VAR;
	    p=&o1;
	    ret=1;
	}
	else {
	    check_delete(&o1);
	    error=SYNTAX_ERROR;
	}
	break;
     case (operator)dollar:
	if (pop_obj(&o1));
	else if (error) return 0;
	
	
	if (o1.type==TYPE_NUM_VAR) {
	    o1.type=TYPE_NUM;
	    o1.num=*VAR[o1.num].num;
	}
	
	if (o1.type==TYPE_NUM) {
	    if (o1.num<-NUMVAR || o1.num>=NUMPARAM) {
		error=OUT_RANGE_ERROR;
		break;
	    }
	    o1.type=TYPE_TXT_VAR;
	    p=&o1;
	    ret=1;
	}
	else {
	    check_delete(&o1);
	    error=SYNTAX_ERROR;
	}
	break;
     case (operator)pre_null:
     case (operator)post_null:
	ret=1;
	break;
     default:
	break;
    }
    
    if (error)
	print_error(error);
    else if (!error && !ret)
	print_error(error=NOT_DONE_ERROR);
    else if (!error && p && push_obj(p))
	;
    else if (error)
	return 0;
    
    return ret;
}

int whichfirst(op1, op2)
operator *op1, *op2;
{
    unsigned p1, p2;
    
    p1=op_list[*op1].priority;
    p2=op_list[*op2].priority;
    if (p1!=p2)
	return p1>p2 ? -1 : 1;
    
    p1 = op_list[*op1].assoc == LEFT;
    return p1 ? -1 : 1;
}

int compare_and_unload(op)
operator *op;
{
    int first=0;
    operator new;
    
    if (error || stk.curr_op<0)
	return 1;
    
    while (stk.curr_op>=0 && pop_op(&new) && !error &&
	   (first = whichfirst(&new, op)) == -1 &&
	   (first = 0, exe_op(&new))
	   );
    
    if (!error)
	if (!first)
	return 1;
    else
	return push_op(&new);
    else
	return 0;
}

int _eval(mindepth)
int mindepth;
{
    operator op;
    object obj;
    char endreached = 0;
    
    for (;;) {
	bzero(&obj, sizeof(obj));
	
	while (*line==' ') line++;
	if (!*line || *line == CMDSEP)
	    endreached = 1;
	
	while (check_operator(LEFT, &op, mindepth) && push_op(&op) &&
	      !endreached) {
	    while (*line==' ') line++;
	    if (!*line || *line == CMDSEP)
		endreached = 1;
	}
	if (error) return 0;
	
	if (!endreached && check_object(&obj) && push_obj(&obj));
	else if (error) return 0;
	
	while (*line==' ') line++;
	if (!*line || *line == CMDSEP)
	    endreached = 1;
	
	while (check_operator(RIGHT, &op, mindepth) && compare_and_unload(&op) &&
	      exe_op(&op) && depth>=mindepth && !endreached) {
	    while (*line==' ')
		line++;
	    if (!*line || *line == CMDSEP)
		endreached = 1; 
	}
	if (error) return 0;
	
	if (endreached || depth < mindepth)
	    break;
	
	if (check_operator(BINARY, &op, mindepth) &&
	    compare_and_unload(&op) && push_op(&op));
	else if (error) return 0;
    }
    return 1;
}

int eval(result, what, printmode)
char **result, **what; int printmode;
{
    int i=0;
    long val;
    ptr txt;
    object res;
    
    if (!result || !*result)
	printmode = PRINT_NOTHING;
    
    error=0;
    stk.curr_obj=stk.curr_op=-1;
    line = *what;
    
    depth = 0;
    (void)_eval(0);
    
    if (!error)
	(void)pop_obj(&res);
    if (error) {
	if (opt_debug)
	    tty_puts("#result not available\n");
    }
    else if (printmode!=PRINT_NOTHING || opt_debug) {
	if (res.type==TYPE_NUM || res.type==TYPE_NUM_VAR) {
	    
	    val = res.type==TYPE_NUM ? res.num : *VAR[res.num].num;
	    
	    if (opt_debug) {
		PRINT_INT ("#result: %ld\n", val);
	    }
	    
	    if (printmode==PRINT_AS_TEXT) {
		sprintf(*result, "%ld", val);
		i = strlen(*result);
	    }
	    else if (printmode==PRINT_AS_NUM) {
		*(long *)*result=val;
		i = sizeof(long);
	    }
	}
	else {
	    txt = res.type==TYPE_TXT ? res.txt : *VAR[res.num].str;
	    if (opt_debug) {
		if (txt) {
		    PRINT_INT ("#result: %.*s\n", ptrlen(txt), ptrdata(txt));
		} else {
		    PRINT_INT ("#empty result\n", ptrlen(txt), ptrdata(txt));
		}
	    }
	    if (printmode!=PRINT_NOTHING) {
		if (txt) {
		    strncpy(*result, ptrdata(txt), ptrlen(txt));
		    (*result)[i=ptrlen(txt)] = '\0';
		} else
		    (*result)[i=0] = '\0';
	    }
	}
	if (printmode!=PRINT_NOTHING)
	    *result+=i;
    }
    *what=line;
    
    if (!error) 
	check_delete(&res);
    else {
	while (stk.curr_obj>=0) {
	    pop_obj(&res);
	    check_delete(&res);
	}
	res.type = 0;
    }
    
    if (res.type==TYPE_TXT_VAR)
	res.type = TYPE_TXT;
    else if (res.type==TYPE_NUM_VAR)
	res.type = TYPE_NUM;
    
    return res.type;
}
